

#include "ruuvi.hxx"
#include <getopt.h>
#include <sys/wait.h>
#include <unistd.h>
#include <chrono>
#include <thread>

/*
 * This program reads the desired measurement data from a RuuviTag bluetooth sensor.
 * 
 * Usage: ruuvirec hcidev  -p Param [-i RuuviTagId]
 * 
 * The first program argument (hcidev) is mandatory. It defines the bluetooth device, e.g. hci0, to be used by the program.
 * 
 * Param argument options:
 * -p Param (string):	Specifies the data to be read from the RuuviTag. Accepted option values:
 *
'i':    tagid
'd':	data format
'h':	humidity
't':	temperature
'p':	pressure
'b':	battery
* 
*  Param options can be combined. For example, Param option -pth means that both temperature and humidity will be read.
* 
* The Tag id option (-i) is optional:
* 
 * -i RuuviTagId (string):	RuuviTag id. If given, only the measurement data originated from this specific RuuviTag is read.
 * 							Lowercase string format shall be used, e.g, -i fc27d1559d42. In case this argument is not given,
 * 							the measurement data from all detected broadcasting RuuviTag devices are read.
 * 
 * 
 *  Returns:				The program prints out (to std::cout) the requested measurement data received from the RuuviTag bluetooth sensor(s).
 * 
 * 
 * NOTE! Root usage (sudo) is required to execute this program, since low level system device (bluetooth) is used.
 * 
 * 
 * 	Author:	Kari Eerola
 * 	Date:	21042020
 * 
 * */

// External variables.
 extern char *optarg;
extern int optind, opterr, optopt;

// Constant definitions.
//
const std::string szOptions = std::string("p:i:");
const std::string szUsage = std::string("Usage: ruuvirec hcidev (e.g, hci0) -p Param [-i RuuviTagId]");

// Globals.
const std::string szTempFile = "./ruuvi.txt";

int main(int argc, char **argv)
{
std::string szRuuviTagId, szParams, szBtDev;
pid_t pid;

// Require the root usage.
if (getuid())
{
std::cout << "Root usage is required to execute this program. Please use 'sudo'." << std::endl;
std::cout << "Exiting..." << std::endl;
exit(EXIT_FAILURE);
}

// Collect arguments.
if (argc < 3)
{
std::cout << szUsage << std::endl;
std::cout << "Exiting..." << std::endl;
exit(EXIT_FAILURE);    
}
szBtDev = argv[1];
int iOption = 0;

while ((iOption = getopt(argc, argv, szOptions.c_str())) != -1) 
{
	switch ((char)iOption) 
   {
	case 'p':
	{
	if (optarg)
	{
	szParams = std::string(optarg);
	if (!Ruuvi::AreParamsSupported(szParams))
	{
	std::cout << "Invalid Ruuvi parameter. Use these:" << std::endl;
	Ruuvi::PrintValidParams();
	exit(EXIT_FAILURE);
	}
	}
	break;
	}
	case 'i':
	{
	if (optarg)
	{
	szRuuviTagId = std::string(optarg);
	}
	break;
	}
	case '?':
	{
	std::cout << "Invalid option. Valid options are -p and -i." << std::endl;
	std::cout << "Parameters are:" << std::endl;
	Ruuvi::PrintValidParams();
	exit(EXIT_FAILURE);
	}
    case ':':
    {
	std::cout << "Missing mandatory -p option." << std::endl;
	std::cout << "Exiting..." << std::endl;
	exit(EXIT_FAILURE);
	}
    default:
    {
	std::cout << szUsage << std::endl;
	std::cout << "Exiting..." << std::endl;
    exit(EXIT_FAILURE);    
    }
}
}
pid = fork();
if (pid < 0)
{
// Failure in fork()
std::cout << "Fork failed!" << std::endl;
std::cout << "Exiting..." << std::endl;
exit(EXIT_FAILURE);
}

if (pid>0)
{
// Parent process. Wait until the child process has
// finished the bluetooth scan.
wait(NULL);
}
else
{
// Child process. Perform bluetooth scan and terminate. The parent
// takes the control from now on.
Ruuvi::Scan_Dump(szBtDev, szTempFile);
exit(EXIT_SUCCESS);
}

// Parent process. Child has terminated.
Ruuvi oRuuvi(szRuuviTagId, szBtDev,szParams);
// Did the child get results?
if (oRuuvi.GetFileSize(szTempFile) > 0)
{
//Yes. Store the result in oRuuvi.
std::ifstream ifs(szTempFile);
std::string result(std::istreambuf_iterator<char>{ifs}, {});
oRuuvi.SetRuuviResult(result);
}
else
{
//No.
std::cout << "Unable to obtain ruuvitag advertising data." << std::endl;
std::cout << "Exiting..." << std::endl;
Ruuvi::RemoveFile(szTempFile);
exit(EXIT_FAILURE);
}

oRuuvi.PrintBroadcastData();
oRuuvi.RemoveFile(szTempFile);
oRuuvi.ReleaseResources();
exit(EXIT_SUCCESS);
}
