#include "ruuvi.hxx"
#include <regex>
#include <bitset> 
#include <bits/stdc++.h>
#include <iomanip>
#include <algorithm>
#include <map>
#include <cctype>



// Some constants:
const std::string szMfgRuuviInnovations = std::string("FF 99 04");
// Command for reading Ruuvi Tag data:
const std::set<char> sSupportedRuuviParams{'i', 'd',  'h', 't', 'p', 'b'};
//
// Supported Ruuvi parameters:
static std::map<char,std::string> mapParams =
{
{'i', "\"tagId\": "},
{'d', "\"DataFormat\": "},
{'h', "\"Humidity\": "},
{'t', "\"Temperature\": "},
{'p', "\"Pressure\": "},
{'b', "\"Battery:\": "}
};

// Helper function. Tests if the given two ruuvi measurements are identical
bool IdenticalMeasurements(ruuviMeasurement m1, ruuviMeasurement m2)
{
return
(m1.szBeaconId.compare(m2.szBeaconId) == 0) &&
(m1.iDataFormat == m2.iDataFormat) &&
(m1.flHumidity == m2.flHumidity) &&
(m1.flTemperature == m2.flTemperature) &&
(m1.siAccelx == m2.siAccelx) &&
(m1.siAccely == m2.siAccely) &&
(m1.siAccelz == m2.siAccelz);
}


//////////////////////////////////////////////////////////////////
//
// Constructor
//
//////////////////////////////////////////////////////////////////
Ruuvi::Ruuvi(const std::string& rszTagId,
			 const std::string& rszBluetoothDevice,
			 const std::string& rszTargetParams)
: mszTagId(rszTagId),
  mszTargetParams(rszTargetParams),
  mszBluetoothDevice(rszBluetoothDevice)
{
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////// //////////////////////
//
// Destructor
//
//////////////////////////////////////////////////////////////////
Ruuvi::~Ruuvi(void)
{
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// RemoveFile
//
// Remove the given file as root user.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::RemoveFile(const std::string& rszFilename)
{
// Remove the file if it exists.
 std::ifstream file(rszFilename);
 if (file)
{
std::string szCmd = "sudo rm ";
    szCmd +=rszFilename;
    system(szCmd.c_str());
}
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// RemoveLinebreaks
//
// Remove all line breaks from a string.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::RemoveLinebreaks( std::string& szStr)
{
szStr.erase(std::remove(szStr.begin(), szStr.end(), '\n'), szStr.end());
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// ReplaceSubstring
//
// Replace  all substring occurrences in a string with another string.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::ReplaceSubstring( std::string& szStr, const std::string& rszSubstr, const std::string& rszReplacestr)
{
szStr.erase(std::remove(szStr.begin(), szStr.end(), '\n'), szStr.end());
szStr = std::regex_replace(szStr, std::regex(rszSubstr), rszReplacestr);
}
//////////////////////////////////////////////////////////////////
//
// ReleaseResources
//
// Release any system resources allocated by this object.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::ReleaseResources(void)
{
// Terminate background processes launced by this object, since they
//
// might still be running.

const std::string szKillScan = "sudo killall -9 hcitool";
const std::string szKillDump = "sudo killall -9 hcidump";
system(szKillScan.c_str());
system(szKillDump.c_str());	
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// AreParamsSupported
//
// Return true in case the given Ruuvi parameter string valid,
// false otherwise.
//
//////////////////////////////////////////////////////////////////
bool Ruuvi::AreParamsSupported( std::string& szParams)
{
bool fSupported = true;
for(std::string::iterator it = szParams.begin(); it != szParams.end(); ++it)
{
if ((sSupportedRuuviParams.count(*it)) <= 0)
  fSupported = false;
}
return fSupported;
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// PrintValidParams
//
// Print the valid Ruuvi parameters to cout.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::PrintValidParams(void)
{
std::map<char,std::string>::iterator it;
it = mapParams.begin();
while (it != mapParams.end())
{
 char key = it->first;
 std::string value = it->second;
 std::cout << key << ": " << value << std::endl;
 it++;
}
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// Change the supplied string to lowercase.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::Tolowercase(std::string& rStr)
{
std::transform(rStr.begin(), rStr.end(), rStr.begin(),[](unsigned char c){ return std::tolower(c); });
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// Return the binary string corresponding to given hex digit.
//
//////////////////////////////////////////////////////////////////
const char* Ruuvi::HexCharToBin(char c)
{
switch(toupper(c))
    {
        case '0': return "0000";
        case '1': return "0001";
        case '2': return "0010";
        case '3': return "0011";
        case '4': return "0100";
        case '5': return "0101";
        case '6': return "0110";
        case '7': return "0111";
        case '8': return "1000";
        case '9': return "1001";
        case 'A': return "1010";
        case 'B': return "1011";
        case 'C': return "1100";
        case 'D': return "1101";
        case 'E': return "1110";
        case 'F': return "1111";
    }
 return NULL;
   
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
// Return binary string corresponding to given hex string.
//
//////////////////////////////////////////////////////////////////
std::string Ruuvi::HexStrToBinStr(const std::string& strHex)
{
std::string strBin;
for(unsigned i = 0; i != strHex.length(); ++i)
  strBin += HexCharToBin(strHex[i]);
  
return strBin;
}













//////////////////////////////////////////////////////////////////
//
// FormBeaconId
//
// Form the beacon id out of the supplied 'hci dump' id text.
//
//////////////////////////////////////////////////////////////////
std::string Ruuvi::FormBeaconId(const std::string& rszHciDumpText)
{
std::string szBeacon;

if (rszHciDumpText.length() == ilBeaconId)
{
std::string str;
for (int i = ilBeaconId -2; i >= 0;)
{
std::string str(rszHciDumpText.substr(i, 2));
str += " ";
szBeacon.append(str);
i -= 3;
}
}
else
{
std::cout << "Unexpected beacon format!" << std::endl;
}
std::string str(szBeacon);
return szBeacon;
}
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//
//
// PrintBroadcastData
//
// Print all detected measurement data to cout.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::PrintBroadcastData(void)
{
for(unsigned int i=0; i < mvbroadcastData.size(); i++)
{
std::string szResult = mvbroadcastData[i];
ruuviMeasurement stRm = mvruuviMeasurements[i];

// Ignore results from "wrong" ruuvitags
if (!mszTagId.empty() && stRm.szBeaconId.compare(mszTagId) != 0)
return;

if (mszTargetParams.find("i") != std::string::npos)
    std::cout << "RuuviTagId: " << stRm.szBeaconId << std::endl; 
if (mszTargetParams.find("d") != std::string::npos)
  std::cout << "DataFormat: " << stRm.iDataFormat << std::endl;  
if (mszTargetParams.find("t") != std::string::npos)
    std::cout << "Temperature: " << stRm.flTemperature << " C" << std::endl;
if (mszTargetParams.find("h") != std::string::npos)
	std::cout << "Humidity: " << stRm.flHumidity << " %" << std::endl;
if (mszTargetParams.find("p") != std::string::npos)   
	std::cout << "Pressure: " << stRm.ulPressure << " Pa" << std::endl;
if (mszTargetParams.find("X") != std::string::npos)
   std::cout << "Accelx: " << stRm.siAccelx << std::endl;
if (mszTargetParams.find("Y") != std::string::npos)
    std::cout << "Accely: " << stRm.siAccely << std::endl;
if (mszTargetParams.find("X") != std::string::npos)
	std::cout << "Accelz: " << stRm.siAccelz << std::endl;
if (mszTargetParams.find("b") != std::string::npos)
	std::cout << "Battery: " << stRm.flBattery << " V" << std::endl;
}
}
//////////////////////////////////////////////////////////////////
//
// Scan_Dump
//
// Perform the bluetooth scan and dump the obtained data in supplied file.
//
//////////////////////////////////////////////////////////////////
void Ruuvi::Scan_Dump(const std::string& rszBluetoothDev, const std::string& rszResultFile)
{
// Form device initialisation (reset) command.
std::string szInitCmd= "sudo hciconfig ";
szInitCmd += rszBluetoothDev;
szInitCmd += " reset";

// Form ble scanning command.
std::string szScanCmd= "sudo hcitool lescan 2>&1 >/dev/null &";

// Form ble dump command.
std::string szDumpCmd= "sudo hcidump --raw 2>&1 >/dev/null > ";
szDumpCmd += rszResultFile;
szDumpCmd += " &";

//Initialise bluetooth.
std::cout << "Initialising bluetooth..." << std::endl;
if (system(szInitCmd.c_str()) != 0)
{
std::cout << "Unable to reset bluetooth device " << rszBluetoothDev << " . Exiting..." << std::endl;
exit(EXIT_FAILURE);
}
// Wait for a while.
std::this_thread::sleep_for( std::chrono::seconds(2));

// Begin BLE scan.
RemoveFile(rszResultFile);
std::cout << "Scanning..." << std::endl;
if (system(szScanCmd.c_str()) != 0)
{
std::cout << "Unable to start bluetooth low energy scan. Exiting..." << std::endl;
exit(EXIT_FAILURE);
}

// Begin BLE dump.
std::cout << "Analysing..." << std::endl;
if (system(szDumpCmd.c_str()) != 0)
{
std::cout << "Unable to perform bluetooth dump. Exiting..." << std::endl;
exit(EXIT_FAILURE);
}

std::cout << "\r ";

// Wait for dump to proceed.
bool fContinueWait = true;
const int iMaxWait = 4;
int i = iMaxWait;
while (fContinueWait && i > 0)
{
std::this_thread::sleep_for( std::chrono::seconds(3));
fContinueWait = GetFileSize(rszResultFile) > 0 ? false : true;
std::this_thread::sleep_for( std::chrono::seconds(3));
i--;
}
if ( i<= 0)
{
std::cout << "No dump results. Exiting..." << std::endl;
exit(EXIT_FAILURE);
}
}
//////////////////////////////////////////////////////
//
// GetFileSize

// Return the size in bytes of the given file.
//
//////////////////////////////////////////////////////////////////
size_t Ruuvi::GetFileSize(const std::string& rszFilename)
{
struct stat st;
size_t filesize = 0;

if(stat(rszFilename.c_str(), &st) == 0)
{
filesize = st.st_size;
}

return filesize;
}
//////////////////////////////////////////////////////////////////
//
// RegExprMatchpoint
//
// Search for the given regular expression (rszRe) in the given
// target string (rszString) and return the position of  the first
// match point. Returns std::string::npos in case of no match.
//
//////////////////////////////////////////////////////////////////
size_t Ruuvi::RegExprMatchpoint(const std::string& rszString, const std::string& rszRe)
{
std::regex rgx(rszRe);
std::smatch match;
size_t pos = std::string::npos;

// Find the matches.
for ( std::sregex_iterator it= std::sregex_iterator(rszString.begin(), rszString.end(), rgx); it != std::sregex_iterator(); it++)
{
std::smatch match = *it;
pos = match.position(0);
}
return pos;
}
//////////////////////////////////////////////////////////////////
//
// SetRuuviResult
//
// Set the ruuvi result into this object.
//
//////////////////////////////////////////////////////////////////
void  Ruuvi::SetRuuviResult(std::string& rszResult)
{
// 1. Store the result as a single string.
mszRuuviResult =rszResult;

std::string szResult(rszResult);
RemoveLinebreaks(szResult);
// 2. Split the result into substrings, each representing a single broadcast data.
const std::string szSearch = "> ";
size_t pos, posStart = 0, posLasthit = 0;
while ((pos = szResult.find(szSearch, posStart)) != std::string::npos)
{
std::string szBroadcast(szResult.substr(posLasthit,pos - posLasthit));
ReplaceSubstring(szBroadcast, "  ", "");
// We deal only with data originated from Ruuvi Innovations devices.
if ((RegExprMatchpoint(szBroadcast, szMfgRuuviInnovations)) != std::string::npos)
{
ruuviMeasurement stRm = GetRuuviMeasurement(szBroadcast);
//Ignore duplicates and data from incorrect ruuvitag device.
if (!RuuviMeasurementExists(stRm) && (mszTagId.empty() || mszTagId == stRm.szBeaconId))
  {
   mvbroadcastData.push_back(szBroadcast);
   mvruuviMeasurements.push_back(stRm);
  }
}
posLasthit = pos;
posStart = pos + szSearch.length();
}
}
//////////////////////////////////////////////////////////////////
//
// GetRuuviMeasurement
//
// Form ruuviMeasurement from the supplied broadcast data.
//
//////////////////////////////////////////////////////////////////
ruuviMeasurement  Ruuvi::GetRuuviMeasurement(const std::string& rszBroadcastData)
{
ruuviMeasurement stRuuviData = {"", 0, 0, 0};

size_t posMfgId = RegExprMatchpoint(rszBroadcastData, szMfgRuuviInnovations);
if (posMfgId == std::string::npos)
{
// Broadcast data not from Ruuvi Innovations.
// Return NULL measurement data.
return stRuuviData;
}
size_t posBeaconId = posMfgId + ioBeaconId;
size_t posDataFormat = posMfgId + ioDataFormat;
std::string szParameterValue = rszBroadcastData.substr(posDataFormat, ilDataFormat);
int iDataFormat = std::stoi(szParameterValue);
stRuuviData.iDataFormat = iDataFormat;

size_t posHumidity,posBattery,posPressure,posAccelx, posAccely,posAccelz;
size_t posTemp1 = posMfgId + ioTemp1;
size_t posTemp2 = posMfgId + ioTemp2;
size_t posTemp = posMfgId + ioTemp5;


if (iDataFormat == 3)
{
posHumidity = posMfgId + ioHumidity;
posPressure = posMfgId + ioPressure;
posAccelx = posMfgId + ioAccelx;
posAccely = posMfgId + ioAccely;
posAccelz = posMfgId + ioAccelz;
posBattery = posMfgId + ioBattery;
}
else if (iDataFormat == 5)
{
posHumidity = posMfgId + ioHumidity5;
posPressure = posMfgId + ioPressure5;
posAccelx = posMfgId + ioAccelx5;
posAccely = posMfgId + ioAccely5;
posAccelz = posMfgId + ioAccelz5;
posBattery = posMfgId + ioBattery5;
}

szParameterValue = rszBroadcastData.substr(posBeaconId, ilBeaconId);
std::string szBeaconId(FormBeaconId(szParameterValue));
ReplaceSubstring(szBeaconId, " ", "");
Tolowercase(szBeaconId);
stRuuviData.szBeaconId = szBeaconId;

if (iDataFormat == 3)
{
szParameterValue = rszBroadcastData.substr(posHumidity, ilHumidity);
int iHumidity= std::stoi(szParameterValue, 0, 16);
stRuuviData.flHumidity = iHumidity * 0.5;

szParameterValue = rszBroadcastData.substr(posTemp1, ilTemp1);
bool fTempNegative = szParameterValue[0] & 0x80 ? true : false;
szParameterValue[0] &= 0x7F;
int iTemp1 = std::stoi(szParameterValue, 0, 16);

szParameterValue = rszBroadcastData.substr(posTemp2, ilTemp2);
int iTemp2 = std::stoi(szParameterValue, 0, 16);
float flTemp = iTemp1 + (0.01 * iTemp2);
if (fTempNegative)
{
flTemp *= -1.0;
}
stRuuviData.flTemperature = flTemp;

szParameterValue = rszBroadcastData.substr(posBattery, ilBattery);
std::string szBattery(szParameterValue);
ReplaceSubstring(szBattery, " ", "");
unsigned long ulBattery = std::stoul(szBattery, 0, 16);
float flBattery = ulBattery/1000.0;
stRuuviData.flBattery = flBattery;
szParameterValue = rszBroadcastData.substr(posPressure, ilPressure);
std::string szPressure(szParameterValue);
ReplaceSubstring(szPressure, " ", "");
unsigned long ulPressure = std::stoul(szPressure, 0, 16);
ulPressure += 50000;
stRuuviData.ulPressure = ulPressure;

szParameterValue = rszBroadcastData.substr(posAccelx, ilAccelx);
std::string szAccelx(szParameterValue);
ReplaceSubstring(szAccelx, " ", "");
int sAccel= std::stoi(szAccelx, 0, 16);
if (sAccel > -16000 && sAccel < 16000)
stRuuviData.siAccelx = sAccel;
szParameterValue = rszBroadcastData.substr(posAccely, ilAccely);
std::string szAccely(szParameterValue);
ReplaceSubstring(szAccely, " ", "");
stRuuviData.siAccely = std::stoi(szAccely, 0, 16);
szParameterValue = rszBroadcastData.substr(posAccelz, ilAccelz);
std::string szAccelz(szParameterValue);
ReplaceSubstring(szAccelz, " ", "");
stRuuviData.siAccelz = std::stoi(szAccelz, 0, 16);
}
else if (iDataFormat == 5)
{
szParameterValue = rszBroadcastData.substr(posHumidity, ilHumidity5);
std::string szHum(szParameterValue);
ReplaceSubstring(szHum, " ", "");
unsigned long ulHum = std::stoul(szHum, 0, 16);
float flHum = ulHum/400.0;
stRuuviData.flHumidity = flHum;

szParameterValue = rszBroadcastData.substr(posTemp, ilTemp5);
std::string szTemp(szParameterValue);
ReplaceSubstring(szTemp, " ", "");
long int lTemp = std::stol(szTemp, 0, 16);
float flTemp = lTemp/200.0;
stRuuviData.flTemperature = flTemp;

szParameterValue = rszBroadcastData.substr(posMfgId + ioPressure5, ilPressure5);
std::string szPressure(szParameterValue);
ReplaceSubstring(szPressure, " ", "");
unsigned long ulPressure = std::stoul(szPressure, 0, 16);
ulPressure += 50000;
stRuuviData.ulPressure = ulPressure;

szParameterValue = rszBroadcastData.substr(posBattery, ilBattery5);
std::string szBattery(szParameterValue);
ReplaceSubstring(szBattery, " ", "");
std::string szVoltage = HexStrToBinStr(szBattery);
// First 11 bits tell the current voltage above 1.6V.
szVoltage = szVoltage.substr(0, 11);
unsigned long  ulBattery = std::stoul(szVoltage, nullptr, 2);
ulBattery += 1600;
float flBattery = ulBattery/1000.0;
stRuuviData.flBattery = flBattery;
szParameterValue = rszBroadcastData.substr(posAccelx, ilAccelx5);
std::string szAccelx(szParameterValue);
ReplaceSubstring(szAccelx, " ", "");
int sAccel= std::stoi(szAccelx, 0, 16);
if (sAccel > -16000 && sAccel < 16000)
stRuuviData.siAccelx = sAccel;
szParameterValue = rszBroadcastData.substr(posAccely, ilAccely5);
std::string szAccely(szParameterValue);
ReplaceSubstring(szAccely, " ", "");
stRuuviData.siAccely = std::stoi(szAccely, 0, 16);
szParameterValue = rszBroadcastData.substr(posAccelz, ilAccelz5);
std::string szAccelz(szParameterValue);
ReplaceSubstring(szAccelz, " ", "");
stRuuviData.siAccelz = std::stoi(szAccelz, 0, 16);

}

return stRuuviData;
}
//////////////////////////////////////////////////////////////////
//
// GetRuuviResult
// Return the ruuvi result.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
std::string Ruuvi::GetRuuviResult(void)
{
return mszRuuviResult;
}
//////////////////////////////////////////////////////

// Return  Ruuvitag id.

//////////////////////////////////////////////////////
std::string Ruuvi::GetRuuviTagId(void)
{
return mszTagId;
}
//////////////////////////////////////////////////////////////////
//
// RuuviMeasurementExists
// Tests if this object already has the given measurement data.
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool Ruuvi::RuuviMeasurementExists(const ruuviMeasurement& stRm)
{
bool fExists = false;
for(unsigned int i=0; i < mvruuviMeasurements.size(); i++)
{
if (IdenticalMeasurements(mvruuviMeasurements[i], stRm))
  {
  fExists = true;
  break;
  }
}

return fExists;
}
//////////////////////////////////////////////////////////////////
